@@ -162,6 +162,8 @@ class Host(object):
         # kernel, QEMU, and/or libvirt. These are determined on demand and
         # memoized by various properties below
         self._supports_amd_sev: ty.Optional[bool] = None
+        self._max_sev_guests: ty.Optional[int] = None
+        self._max_sev_es_guests: ty.Optional[int] = None
         self._supports_uefi: ty.Optional[bool] = None
         self._supports_secure_boot: ty.Optional[bool] = None
 
@@ -1252,6 +1254,59 @@ class Host(object):
         """
         return self.get_connection().nodeDeviceLookupByName(name)
 
+    def device_create(self, conf):
+        """Create a node device from specified device XML
+
+        This creates the device as transient.
+
+        :param conf: A LibvirtConfigObject of the device to create
+
+        :returns: a virNodeDevice instance if successful, else None
+        """
+        device_xml = conf.to_xml()
+        return self.get_connection().nodeDeviceCreateXML(device_xml, flags=0)
+
+    def device_define(self, conf):
+        """Define a node device from specified device XML
+
+        This defines the device to make it persistent.
+
+        :param conf: A LibvirtConfigObject of the device to create
+
+        :returns: a virNodeDevice instance if successful, else None
+        """
+        device_xml = conf.to_xml()
+        return self.get_connection().nodeDeviceDefineXML(device_xml, flags=0)
+
+    def device_start(self, dev):
+        """Start a defined node device
+
+        :param dev: The virNodeDevice instance to start
+        """
+        # extra flags; not used yet, so callers should always pass 0
+        #   https://libvirt.org/html/libvirt-libvirt-nodedev.html
+        flags = 0
+        result = dev.create(flags)
+        if result == -1:
+            msg = f'Failed to start node device {dev.name()}'
+            raise exception.InternalError(_(msg))
+
+    def device_set_autostart(self, dev, autostart=True):
+        """Set a node device to automatically start when the host boots
+
+        This can set whether the node device should automatically start when
+        the host machine boots or when the parent device becomes available.
+
+        :param dev: The virNodeDevice instance to set the autostart value
+        :param autostart: Whether to set the device to automatically start
+        """
+        result = dev.setAutostart(autostart=autostart)
+        if result == -1:
+            msg = (
+                f'Failed to set autostart to {autostart} for node device '
+                f'{dev.name()}')
+            raise exception.InternalError(_(msg))
+
     def _get_pcinet_info(
         self,
         dev: 'libvirt.virNodeDevice',
@@ -1609,7 +1664,7 @@ class Host(object):
         """Lookup devices.
 
         :param flags: a bitmask of flags to filter the returned devices.
-        :returns: a list of virNodeDevice xml strings.
+        :returns: a list of virNodeDevice instances.
         """
         try:
             alldevs = [
@@ -1775,7 +1830,7 @@ class Host(object):
             return self._supports_secure_boot
 
         # we only check the host architecture since the libvirt driver doesn't
-        # truely support non-host architectures currently
+        # truly support non-host architectures currently
         arch = self.get_capabilities().host.cpu.arch
         domain_caps = self.get_domain_capabilities()
         for machine_type in domain_caps[arch]:
@@ -1793,6 +1848,57 @@ class Host(object):
         self._supports_secure_boot = False
         return False
 
+    @property
+    def supports_vtpm(self) -> ty.Optional[bool]:
+        # we only check the host architecture and the first machine type
+        # because vtpm support is independent from cpu architecture
+        arch = self.get_capabilities().host.cpu.arch
+        domain_caps = self.get_domain_capabilities()
+        for machine_type in domain_caps[arch]:
+            _tpm = domain_caps[arch][machine_type].devices.tpm
+            # TODO(tkajinam): Remove this once libvirt >= 8.0.0 is required
+            if _tpm is None:
+                return None
+            return (_tpm.supported and 'emulator' in _tpm.backend_models)
+        # safe guard
+        return False
+
+    @property
+    def tpm_versions(self) -> ty.Optional[ty.List[str]]:
+        # we only check the host architecture and the first machine type
+        # because vtpm support is independent from cpu architecture
+        arch = self.get_capabilities().host.cpu.arch
+        domain_caps = self.get_domain_capabilities()
+        for machine_type in domain_caps[arch]:
+            _tpm = domain_caps[arch][machine_type].devices.tpm
+            # TODO(tkajinam): Remove first check once libvirt >= 8.0.0 is
+            # required
+            # TODO(tkajinam): Remove second check once libvirt >= 8.6.0 is
+            # required
+            if _tpm is None or _tpm.backend_versions is None:
+                return None
+            return _tpm.backend_versions
+        # safe guard
+        return []
+
+    @property
+    def tpm_models(self) -> ty.Optional[ty.List[str]]:
+        # we only check the host architecture and the first machine type
+        # because vtpm support is independent from cpu architecture
+        arch = self.get_capabilities().host.cpu.arch
+        domain_caps = self.get_domain_capabilities()
+        for machine_type in domain_caps[arch]:
+            _tpm = domain_caps[arch][machine_type].devices.tpm
+            # TODO(tkajinam): Remove first check once libvirt >= 8.0.0 is
+            # required
+            # TODO(tkajinam): Remove second check once libvirt >= 8.6.0 is
+            # required
+            if _tpm is None or _tpm.models is None:
+                return None
+            return _tpm.models
+        # safe guard
+        return []
+
     def _kernel_supports_amd_sev(self) -> bool:
         if not os.path.exists(SEV_KERNEL_PARAM_FILE):
             LOG.debug("%s does not exist", SEV_KERNEL_PARAM_FILE)
@@ -1839,11 +1945,29 @@ class Host(object):
                     if feature_is_sev and feature.supported:
                         LOG.info("AMD SEV support detected")
                         self._supports_amd_sev = True
+                        self._max_sev_guests = feature.max_guests
+                        self._max_sev_es_guests = feature.max_es_guests
                         return self._supports_amd_sev
 
         LOG.debug("No AMD SEV support detected for any (arch, machine_type)")
         return self._supports_amd_sev
 
+    @property
+    def max_sev_guests(self) -> ty.Optional[int]:
+        """Determine maximum number of guests with AMD SEV.
+        """
+        if not self.supports_amd_sev:
+            return None
+        return self._max_sev_guests
+
+    @property
+    def max_sev_es_guests(self) -> ty.Optional[int]:
+        """Determine maximum number of guests with AMD SEV-ES.
+        """
+        if not self.supports_amd_sev:
+            return None
+        return self._max_sev_es_guests
+
     @property
     def supports_remote_managed_ports(self) -> bool:
         """Determine if the host supports remote managed ports.
